/* polyathlon.c 
// program by Dale Heatherington  started April 23, 2009

Version 1.1  May 15 2009
This one has Maze solving by right and left hand rule

Compiled with AVR-GCC on Apple Mac Pro
Note: CPU type defined in Makefile but it must be a ATMEGA8535
      Brownout protection mode must be set in the low fuse byte to prevent EEPROM corruption.
	  This is done in the makefile in the "pgm" section for the avrdude programmer
	  
	  Initially the line sensor photocell array must be calibrated on a white surface
	  by setting the mode switch to 8 and turning on power.  Values are stored in EEPROM.
	  

This is a work in progress is may be incomplete or infested with bugs.
It has been tweeked to work with the hardware and sensors available and will 
probably not function on any other hardware.  It is presented for educational
purposes.



;;----------------------------------------------------------------------
;;This program is free software; you can redistribute it and/or modify  
;;it under the terms of the GNU General Public License as published by      
;;the Free Software Foundation; either version 2 of the License, or         
;;(at your option) any later version.                                       
;;
;;This program is distributed in the hope that it will be useful,       
;;but WITHOUT ANY WARRANTY; without even the implied warranty of            
;;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             
;;GNU General Public License for more details.                              
                                                                          
;;You should have received a copy of the GNU General Public License     
;;along with this program; if not, write to the Free Software               
;;Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA
;;------------------------------------------------------------------------

*/
 



#include <avr/io.h>
#include <stdio.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#include <avr/eeprom.h>
#include <stdlib.h>
#include <string.h>
#include <avr/wdt.h>

/* I/O and bit operation macros that have were 
   removed from late versions of avr-libc */

#define  _BV(bit)   (1 << (bit))
#define  inb(sfr)   _SFR_BYTE(sfr)
#define  inw(sfr)   _SFR_WORD(sfr)
#define  outb(sfr, val)   (_SFR_BYTE(sfr) = (val))
#define  outw(sfr, val)   (_SFR_WORD(sfr) = (val))
#define  cbi(sfr, bit)   (_SFR_BYTE(sfr) &= ~_BV(bit))
#define  sbi(sfr, bit)   (_SFR_BYTE(sfr) |= _BV(bit))

#ifndef bit_is_set
//#define  bit_is_set(sfr, bit)   (inb(sfr) & _BV(bit))
#endif

#ifndef bit_is_clear
//#define  bit_is_clear(sfr, bit)   (!(inb(sfr) & _BV(bit)))
#endif

#define  loop_until_bit_is_set(sfr, bit)   do { } while (bit_is_clear(sfr, bit))
#define  loop_until_bit_is_clear(sfr, bit)   do { } while (bit_is_set(sfr, bit))
#define  outp(val, sfr)   outb(sfr, val)
#define  inp(sfr)   inb(sfr)
#define  BV(bit)   _BV(bit)
/*---------------------------------------------------------------------------------*/


#define F_CPU    16000000       /*  16 mhz, crystal frequency of CPU */

#define WDTO_250MS 4        /* Watchdog timer value = 250ms */
#define UCHAR uint8_t
#define TRUE 1
#define FALSE 0

#define MODE_STOP 0
#define REV 0
#define FWD 1
#define CW 0
#define CCW 1

#define LEFT 0
#define RIGHT 1
#define INTERSECTION 2
#define NONE 3

#define MAX_FWD_SPEED -127
#define MAX_REV_SPEED 127

#define MAXTIME 180

//Threshold for IR edge detectors used as line detectors
//White is above this number, black below.
#define ET 0xe0



//Counts per inch integer (actually 36.9)
#define CPI 37


/* Define string constants in FLASH program memory */
const char testquad[] PROGMEM = "TestQuad";
const char linefoll[] PROGMEM = "LineFoll";
const char beaconfinder[]  PROGMEM = "Beacon";
const char deadrec[]  PROGMEM = "DeadRec";
const char bulldoz[]  PROGMEM = "BullDoz";
const char linemaze[] PROGMEM = "LineMaze";
const char whitelevel[] PROGMEM = "WhiteLvl";



/*--- Motor control variables and defs ----- */

//Bit in PORTD that enables drive motor h-bridges

int r_speed,r_motor;
int l_speed,l_motor;

int speed;
char dir;

UCHAR m_current;





volatile UCHAR timer_mstop;
UCHAR prescale ;
UCHAR prescale100;
UCHAR prescale1sec;
UCHAR prescaleMinute;

UCHAR timer1sec;
volatile unsigned int minutes;
volatile int run_seconds,timer_trip, timer_move;
volatile int timer_led_yellow, timer10ms, timer100us, timer_overcurrent, timer_reverse, timer_delay;
volatile UCHAR rnd;

//Quadrature encoder data
volatile unsigned long odo_left, odo_right;
volatile  long trip_odo_right, trip_odo_left;
volatile long line_history_odo_right, line_history_odo_left;
volatile int speed_left, speed_right;
volatile UCHAR dir_left, dir_right;


char *tx_buf;
volatile UCHAR tx_ptr,tx_ctr ;
volatile UCHAR txstate;


UCHAR debug[6];  	//Debug byte array - put data to be displayed on LCD here.

UCHAR last_sec;

UCHAR trip_counter;
UCHAR stop;
UCHAR sel_sw;
UCHAR edge_left, edge_right;  //Table edge detector analog values 0..255
int avg_contrast[4];

UCHAR   pot, motor_current;


struct Line {	
	int centroid;
	int contrast;
	UCHAR brightness;
	UCHAR present;
	UCHAR mean;
	UCHAR map;
	UCHAR width;
	UCHAR data[9];  //last byte is checksum for eeprom storage
};

struct Line line;

struct Beacon {
	int centroid;	
	UCHAR brightness;
	UCHAR left;
	UCHAR right;
	
	
};

struct Beacon beacon;


struct IR {
	int centroid;
	UCHAR brightness;
	UCHAR contrast;
	UCHAR mean;
	UCHAR map;
	UCHAR data[6];
};

struct IR activeIR;
	

char sbuf[24];  //LCD bug tx buffer;




#define OBJ_HIT 4
#define OBJ_AHEAD 3
#define OBJ_LEFT 2
#define OBJ_RIGHT 1
#define OBJ_NONE 0

// Analog to digital converter input defs
#define AD_LINE_FOLLOW 7
#define AD_POT  6
#define AD_EDGE_RIGHT 4
#define AD_EDGE_LEFT 5
#define AD_BEACON_RIGHT 3
#define AD_BEACON_LEFT 2
#define AD_IR_PULSE  1
#define AD_MOTOR_CURRENT 0


UCHAR irsensor[8];
//Define Active IR LEDs in sensor irsensor[] array
#define REV_RIGHT 2
#define REV_LEFT 3
#define FWD_LEFT 1
#define FWD_RIGHT 0





UCHAR over_current(void);
UCHAR IRbumper(UCHAR);
int getContrast(UCHAR , UCHAR*, UCHAR*  );
UCHAR getMap(UCHAR ,  UCHAR*, UCHAR );




//-----------------------------------------------------------------
//------------------------------------------------------------------

//Convert 4 bit value to ascii hex format 0..f

char nibble(UCHAR c)
{	
	c = c & 0x0f;
	if(c < 10) return c + 0x30;   	//0..9
	return c + 0x37;		//A..F
}
	
//Convert 8 bit value to ascii string hex format 00..ff in buffer *cp
//A NULL will be appended to the end of the 2 digit string.
//Supplied buffer must be at least 3 bytes long
void bin2hex(char *cp, UCHAR i)
{  UCHAR t;

	t = i >> 4;
	cp[0] = nibble(t);
	t = i & 0x0f;
	cp[1] = nibble(t);
	cp[2] = 0x00;
		
}
//-------------------------------------------------------------------------


int hard_limit_range(int x, int lim_hi, int lim_lo)
{
   if(x < lim_lo) x = lim_lo; 
   if(x > lim_hi) x = lim_hi; 
   return x;

}
//--------------------------------------------------------------------------
/* Read the 10 bit analog to digital converter */
/*
uint16_t get_adc(UCHAR chan)
{  
    uint16_t result;
	ADMUX = chan;	//Select channel
	
    sbi(ADCSRA,ADSC);               //Start conversion 	
    while( bit_is_clear(ADCSRA,ADIF)); //Wait for ADC output registers to be loaded
    cli();   
	result = ADCW;
    sei();
    return result;       //Return 16 bit value
}
*/
//---------------------------------------------------------------------------

/* Read 10 bit AD converter but return 
   only 8 bits discarding low 2 bits. */

uint8_t get_adc8(UCHAR chan)
{  
    uint16_t result;
   
    ADMUX = chan;					//Select channel
    sbi(ADCSRA,ADSC);                //Start conversion     
    while( bit_is_clear(ADCSRA,ADIF)); //Wait for ADC output registers to be loaded
    cli();   
	result = ADCW;			//Get 10 bit value
    sei();
    return result >> 2;     //Return 8 bit value
}
//---------------------------------------------------------------------------
//Compute simple checksum of array of unsigned bytes.
UCHAR checksum(UCHAR *p, int n)
{   int i;
    UCHAR cksum = 0;
	for(i=0;i<n;i++) cksum += p[i] ;
	return cksum;
}


//---------------------------------------------------------------------------
//Read an active IR seonsor
//Inputs: adr = address of IR LED 0..7
//
UCHAR readActiveIR(UCHAR adr)
{	unsigned int ctr, result;
	adr = adr & 7;	
	
	if(adr & 1) sbi(PORTC,0); else cbi(PORTC,0);
	if(adr & 2) sbi(PORTC,1); else cbi(PORTC,1);
	if(adr & 4) sbi(PORTC,2); else cbi(PORTC,2);
		
	for(ctr=0;ctr<77;ctr++){cbi(PORTC,3);}		//Enable IR LED and wait 30uS
		
		ADMUX = AD_IR_PULSE;
		sbi(ADCSRA,ADSC);						//Start conversion 
		for(ctr=0;ctr<51;ctr++){cbi(PORTC,3);}  //Wait about 20uS for sample and hold
		sbi(PORTC,3);							//IR pulse off
		while( bit_is_clear(ADCSRA,ADIF));		//Wait for ADC output registers to be loaded
		cli();   
		result = ADCW;							//Get 10 bit value
		sei();
		return (UCHAR)(result >> 2)	;			// Return 8 bit result
		

}

//---------------------------------------------------------------------------
int getIRcentroid()
{
	UCHAR i ;
	unsigned int low, sumA, sumB;
	int centroid;
	
	low = 255;
	for( i=0;i<6;i++){		//Find the least bright reading  and store value in "low"
		if(activeIR.data[i] < low) low = activeIR.data[i];
	}
	
	sumA = sumB = 0;
	for( i=0;i<6;i++){				//Compute sums
		sumB += (activeIR.data[i] - low) * (i + 1);
		sumA += activeIR.data[i] - low;
	}
	
	centroid = (sumB * 16) / (sumA/2) ; //Compute and scale centroid
	centroid = centroid - 112;
	
	return centroid;
	
	
	
}


//----------------------------------------------------------------------------

void refreshIRdata()
{
	int i,x;
	UCHAR max = 0;
	
	UCHAR bias = get_adc8(AD_IR_PULSE);  //Get IR photodiode level for ambiant light
	
	for(i=0;i<6;i++){
		x = readActiveIR(i) - bias;  //Read all 6 active IR sensors and subtract bias
		if(x < 0) x = 0;
		activeIR.data[i] = (UCHAR)x;
		
	}
	
	for(i=0;i<6;i++){
		if(activeIR.data[i] > max) max = activeIR.data[i]; //Find max brightness
	}
	
	activeIR.brightness = max;
	activeIR.centroid = getIRcentroid();
	activeIR.contrast = getContrast(6,activeIR.data,&activeIR.mean);
	activeIR.map = getMap(6,activeIR.data,activeIR.mean);
	
	
	
	
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
void readLinePhotocells()
{
	UCHAR adr;
	int x;
	
	eeprom_read_block((void*)&line.data, (const void*)0, 9);  //Read back stored white level reverences
	
	if(checksum(line.data,8) != line.data[8]) {
		eeprom_read_block((void*)&line.data, (const void*)0, 9); //Read spare values if main corrupted
		
		}
	
	for(adr=0;adr<8;adr++){
		if(adr & 1) sbi(PORTC,0); else cbi(PORTC,0);
		if(adr & 2) sbi(PORTC,1); else cbi(PORTC,1);
		if(adr & 4) sbi(PORTC,2); else cbi(PORTC,2);
		
		x = (int)get_adc8(AD_LINE_FOLLOW) - (int)line.data[adr]; //Subtract white level bias
		if(x < 0) x = 0;									//Do not allow negative values
		
		
		line.data[adr] = (UCHAR)(x & 255);  //store corrected value
		
		
	}
	
}

//-------------------------------------------------------------------------
//Compute centroid.  Assumes 8 values of brightness are in line[] array.
//
int getLineCentroid()
{   UCHAR i ;
	unsigned int low, sumA, sumB;
	int centroid;
	
	low = 255;
	for( i=0;i<8;i++){		//Find the brightest photocell and store value in "low"
		if(line.data[i] < low) low = line.data[i];
	}
	
	sumA = sumB = 0;
	for( i=0;i<8;i++){				//Compute sums
		sumB += (line.data[i] - low) * (i + 1);
		sumA += line.data[i] - low;
	}
	
	centroid = (sumB * 16) / (sumA >> 1) ; //Compute and scale centroid
	centroid = centroid - 144;
	line.centroid  = centroid;
	
	return centroid;
}


//-----------------------------------------------------------------------------
		
//-----------------------------------------------------------------------------
//Cpompute and return contrast and mean.
int getContrast(UCHAR n, UCHAR *data, UCHAR *mean )
{
	UCHAR high,low,i;
	unsigned int contrast;
	high = 0;
	low = 255;
	
	for(i=0;i<n;i++){
		if(data[i] > high) high = data[i];
		if(data[i] < low) low = data[i];
	}
	
	UCHAR sliceLevel = high - ((high-low) / 2); //Find level half way between high and low
	contrast = high - low;
		
	*mean = sliceLevel;
	
	return contrast;
	
}
//---------------------------------------------------------------------------


UCHAR getMap(UCHAR n, UCHAR* data, UCHAR mean)    //Photocells reading over the mean value 
						//are black and the assgned bit in "map" is set.
{						//Left most photocell is bit 7, right most is bit 0.
	UCHAR map = 0;
	UCHAR i;
	
	for(i=0;i<n;i++){
		map = map >> 1;
		if(data[i] > mean) map = map | 0x80;
	}
	
	
	return map;
	
}

//---------------------------------------------------------------------------------

	
void refreshLineData()
{   int i, b;
	UCHAR c;
	readLinePhotocells();
	line.contrast = getContrast(8,line.data,&line.mean);	
	line.map = getMap(8,line.data,line.mean);
	
	line.centroid = getLineCentroid();
	
	line.present = FALSE;
	for(i=0;i < 8;i++){
		if(line.data[i] >= 10) line.present = TRUE;
		}
		
	if(line.present == FALSE) {
		line.map = 0;
		line.centroid = 0;
		}
		
	line.width = 0;
	c = 1;
	for(i=0;i<8;i++){
		if(line.map & c ) line.width++;		//count the 1's in line.map to get width
		c = c << 1;		 
		
	}
		
	
	b = 0;
	for(i=0;i<8;i++) b += line.data[i];  //Sum all the photocells
	b = b / 8;  //get average
	line.brightness = 255 - b; //Invert to make brightness a positive function, 0-255

	
}

//------------------------------------------------------------------------------

int getBeaconCentroid()
{   
	unsigned int  sumA, sumB;
	int centroid;
	
	
	//Compute sums
	sumB = beacon.left + (beacon.right  * 2);    //Max = 255 + (255*2) = 755
	sumA = beacon.left + beacon.right;			//Max = 255+255= 510
	
	
	centroid = (sumB * 80)  / sumA ; //Compute and scale centroid. Keep (sumB * X) under 0xffff.
	centroid = centroid - 120;
		
	return centroid;
}

//------------------------------------------------------------------------------
void getBeacon()
{
	
beacon.left = 255 - get_adc8(AD_BEACON_LEFT);
beacon.right = 255 - get_adc8(AD_BEACON_RIGHT);
beacon.brightness = (beacon.right + beacon.left) / 2;
beacon.centroid = getBeaconCentroid();

	
}
	
	

//-----------------------------------------------------------------------------




void delay_100us(uint8_t count)
{
  while(count){
    outp(256 - ((F_CPU/80000) -1), TCNT0);
    outp(BV(TOV0), TIFR);
    while(!(inp(TIFR) & BV(TOV0)));   
    count--;
  }
}


void delay_10ms(uint8_t count)
{
  while(count){
    delay_100us(100);
    count--;
  }
}

//---------------------------------------------------------------------
// Waste time.
// t is in 10 ms intervals
void delay(int t)
{
	timer_delay = t;
	while(timer_delay > 0);
	
}
	

//----------------------------------------------------------------------

/* Adjust motor speeds.  */

#define CURRENT_LIMIT 32
//current limiter
/* 0.27 ohm motor current sampling resistor provideds .27 volts per amp.
   18.5 AD converter units per amp.
   Value of 32 is 1.73 amps (sum of both motors)

*/

void set_motor_speed(int l_speed, int r_speed){
      UCHAR hbridge_enable;
      
      int spd;

      UCHAR lm_pwm, rm_pwm;
	  
				
     
      if((l_speed != 0) || (r_speed != 0)) timer_mstop = 40;  //400ms breaking when speed goes to zero
     
      /* Right motor PWM control */
       
       spd = r_speed ;

       spd = hard_limit_range(spd,125,-126);  //Limit range from +125 to -126 so h-bridge bootstrap always has pulses
       rm_pwm = (UCHAR)spd + 128;
       outp(0,OCR1BH);                      //set right PWM
       outp(rm_pwm,OCR1BL);


       /* Left motor PWM control */   
       
       spd = l_speed;

       spd = hard_limit_range(spd,125,-126);
       lm_pwm = (UCHAR)spd + 128; 
       outp(0,OCR1AH);                      //set left PWM
       outp(lm_pwm,OCR1AL);



       /* H-Bridge enable control */

       hbridge_enable = 1;      //Assume h-bridge will be enabled
                                //Then look for reasons to disable it below...

         if (m_current > CURRENT_LIMIT){
              hbridge_enable = 0;  // Limit drive motor current to 1000 ma
         }

         if ((l_speed == 0)
             && (r_speed == 0)
             && (timer_mstop == 0)) hbridge_enable = 0;   //Disable H-Bridge if bot is stopped
	
	if(hbridge_enable == 0) cbi(PORTC,4); else sbi(PORTC,4);

        

 }


//-----------------------------------------------------------------------------


   /* send debug data to LCD display*/
  
void writelcd(char *cp){
      
	while(txstate != 0);	//Wait until UART not busy      
	tx_buf = cp;
	txstate = 4;         	//Tell scheduler it's ready to transmit
	sbi(UCSRB, TXCIE); 	//Enable TX UART interrupts     
	cbi(UCSRA,TXC);		//Set transmit complete bit
	outp(0x12,UDR);       // Send a cursor to home code char to get things started 
   
      
   }

//----------------------------------------------------------------------
   void refreshDebug()
   {	int i,j;
   
   
   for(i=0,j=0;i<6;i++){
	   bin2hex(sbuf+j,debug[i]);  //Convert all 6 debug values to ascii hex in the sbuf array
	   sbuf[j+2] = ' ';
	   j = j+3;
           if(i == 2) {        //Next line on LCD display after 8 chars.
             sbuf[j++] = '\r';
             sbuf[j++] = '\n';
           }
   }
   writelcd(sbuf);
   }


//----------------------------------------------------------------------

char debugReady()
{
	if(txstate != 0) return 0;
	else return 1;
}
	
//----------------------------------------------------------------------

// Wait t*10ms and display time remaining on LCD display

void delay_start(int t, PGM_P s){
	int tenths;
	
	timer10ms = t;	
		
	while(timer10ms > 0 ){		//Wait here for t*10ms  and show time remaining on LCD
	  
	  tenths = timer10ms / 10;   //Converts hundredths to tenths.
	  
	  if(txstate == 0){
		itoa(tenths,sbuf,10);		//Convert integer to decimal string
		if(tenths > 9){
				sbuf[2] = sbuf[1];
				sbuf[1] = '.';		//Insert decimal point
							
			}
			else
			{
				sbuf[2] = sbuf[0];
				sbuf[1] = '.';
				sbuf[0] = '0';		//Insert leading zero and decimal point
				
			}
		 
		  sbuf[3] = '\r';
		  sbuf[4] = '\n';
		  strcpy_P(&sbuf[5], s);  //Copy Title string from Program memory to ram buffer
		  
		  writelcd(sbuf);			//Write to LCD display
		
		}
	}
	  
  }


//----------------------------------------------------------------------


/* UART Transmit interrupt handler */  
SIGNAL(SIG_UART_TRANS)
{ 
   

      switch(txstate) {
      case 0:  	break;
      case 1:	//unused
      case 2:	//unused
      case 3: 	//unused
      case 4: 	tx_ptr = 0;
         	tx_ctr = strlen(tx_buf); 
         	txstate = 5;
		//fall through...

      case 5:  	outp(tx_buf[tx_ptr++],UDR);  //Now send buffer contents
            	tx_ctr--;
            	if (tx_ctr == 0){ 
               	   txstate = 6;
               
            	}
         	
            break;

      case 6:  
               txstate = 0;
	       cbi(UCSRB,TXCIE);
	       sbi(UCSRA,TXC);    //Set transmit complete bit
	
               break;



      }

      
}

//----------------------------------------------------------------------------

/* Come here 10,000 times per second.   */
SIGNAL(SIG_OUTPUT_COMPARE2)
{

		
		
		rnd++ ;
		rnd = rnd ^ beacon.left;   //Make a random number
		
		if(timer100us > 0) timer100us--;

		if(--prescale == 0){      //This executes 100 times per sec.
			prescale = 100;
         
			if(prescale1sec > 0) prescale1sec--;
			if(timer10ms > 0) timer10ms--;
			if(timer_trip > 0) timer_trip--;
			if(timer_mstop > 0) timer_mstop--;
			if(timer_overcurrent > 0) timer_overcurrent--;
			if(timer_reverse > 0) timer_reverse--;
			if(timer_move > 0) timer_move--;
			if(timer_delay > 0) timer_delay--;
                 
			}

      
		if(prescale1sec == 0){  //This executes once per second
			prescale1sec = 100;
			
			if(prescaleMinute > 0) prescaleMinute--;
			run_seconds++;
			
        
			}

		if(prescaleMinute == 0){ //This executes once per minute
			prescaleMinute = 60;
			minutes++;        
			}

		
 
    /* End interrupt code */
    

}

//-------------------------------------------------------------------------

/* Quadrature encoder interrupt handlers */

/*  36.8  counts per inch  or 0.0272 inch per count  */

//LEFT side quadrature clock
SIGNAL(INT0_vect)
{
	if(PIND & 0x80) dir_left = FWD; else dir_left = REV;
	if(dir_left == FWD){
		odo_left++; 
		trip_odo_left++;
	}else {
		odo_left--;
		trip_odo_left--;
	}
	
}

//Right side Quadrature clock
SIGNAL(INT1_vect)
{
	if(PIND & 0x40) dir_right = FWD; else dir_right = REV;  
	if(dir_right == FWD){
		odo_right++; 
		trip_odo_right++;
	}else {
		odo_right--;
		trip_odo_right--;
	}
	
	
}


//------------------------------------------------------------------------
void refresh_sensors(void)
{

 	pot = get_adc8(AD_POT);	
	motor_current = get_adc8(AD_MOTOR_CURRENT);
	edge_left = get_adc8(AD_EDGE_LEFT);
	edge_right = get_adc8(AD_EDGE_RIGHT);
	getBeacon();
	
	
		
}	
	


//--------------------------------------------------------------------------

//Returns true if motors have been pulling excessive current for too long
UCHAR over_current(void)
{

	motor_current = get_adc8(AD_MOTOR_CURRENT);
	
	if (motor_current < 0x22) timer_overcurrent = 60;  //Reset timer to 600ms if no overcurrent
	
	if ((timer_overcurrent == 0) && (motor_current >= 0x22)) return 1;
	
	return 0;
	
	}






//-------------------------------------------------------------------------
//Rotate in place until bot faces  beacon
UCHAR face_beacon()
{	UCHAR done = 0;
	l_speed = -100;
	r_speed = 100;
	set_motor_speed(l_speed, r_speed);	//start rotating
	timer10ms = 175;
	while((done == 0) && (run_seconds < MAXTIME)){
		refresh_sensors();	
		if((abs(beacon.centroid) < 3) && (beacon.brightness > 0x50)) done = 1;
		if(timer10ms == 0) done = 2;
	}
	
	return done;

}


//------------------------------------------------------------------------------

void rotateToLine(int spd, UCHAR dir)
{
	refreshLineData();
	if( dir == CCW) set_motor_speed(-spd,spd);
		else set_motor_speed(spd,-spd);
	while(line.map == 0) refreshLineData();;
	while(abs(line.centroid) > 10) refreshLineData(); 

}

void rotateOffLine(int spd, UCHAR dir)
{

	refreshLineData();
	if( dir == CCW) set_motor_speed(-spd,spd);
		else set_motor_speed(spd,-spd);
	while(line.map > 0) refreshLineData();;
	

}


//------------------------------------------------------------------------------
/*
void odometer_debug()
{
	if(debugReady()){
	debug[0] = (trip_odo_right >> 16) & 255;
	debug[1] = (trip_odo_right >> 8) & 255;
	debug[2] = (trip_odo_right & 255);
	debug[3] = (trip_odo_left >> 16) & 255;
	debug[4] = (trip_odo_left >> 8) & 255;
	debug[5] = (trip_odo_left & 255);
	refreshDebug();
    }
}
*/	
//------------------------------------------------------------------------------
void driveStraight(int distance, int speed)
{   long dcnt;
	long d;
	int spd;
	trip_odo_left = trip_odo_right = 0;
	dcnt  = CPI * distance;   //d now has odometer counts required
	spd = speed;
	while(trip_odo_left < dcnt){
		if((dcnt - trip_odo_left) < 200) spd = speed / 2;
		
		refresh_sensors();
		
		d = trip_odo_right - trip_odo_left;
		if(d > 5) d = 5;    //Limit odometer difference range to +/- 5
		if(d < -5) d = -5;
		
				
		l_speed = spd + (d * 8);    //Correct steering based on odometer count difference.
		r_speed = spd - (d * 8);
		set_motor_speed(l_speed, r_speed);
		/*
		odometer_debug();
		 */
	}
	
	l_speed = r_speed = 0;
	set_motor_speed(l_speed, r_speed);
		
		
	
}

//----------------------------------------------------------------------------
//Wheels are 7.875 inches apart.
//To turn 360 degrees requires they each travel the circumference which is pi*d = 24.74 inches.
//with 36.9 quadrature counts per inch they need to travel 36.9 * 24.74 = 912.9 counts
// or 2.536 counts per degree.
#define C360 = 913
void rotate(int degrees, int speed, char dir, char option)
{   
	
	long dcnt;
	
	int spd;
	
	
	dcnt = (degrees * 2) + ((degrees * 5) / 10);   //degrees * 2.5
	
		
	spd = speed;
	
	if(dir == CW){				//Set direction to rotate  CW or CCW
		l_speed = spd;
		r_speed = -spd;
		trip_odo_left = -dcnt;
	    trip_odo_right = dcnt;
		} else {
			l_speed = -spd;
			r_speed = spd;
			trip_odo_left =  dcnt;
			trip_odo_right = -dcnt;
			
			}
		
		timer10ms = (degrees * 3) / 4;		//Allow 7.5ms time per degree for turn
		if(degrees < 30) timer10ms = 30;  //300ms minimum
		
		while(timer10ms > 0){
			refresh_sensors();	
			refreshIRdata();	
			if(option == 1){
				refreshIRdata();
				if((activeIR.brightness > 0x18) && (abs(activeIR.centroid) < 10)){
					timer10ms = 0;  //Quit now if object in front of bot.
					}
			}
			
							
			
			//Now let the trip odometer error term totally control the motors like a servo.
			//Eventually the trip odomenter errors will become zero and the motors will stop.
			
			 l_speed = -trip_odo_left * 8;
			l_speed = hard_limit_range(l_speed,spd,-spd);
			 r_speed = -trip_odo_right * 8;
			r_speed = hard_limit_range(r_speed,spd,-spd);
			
			//if((l_speed == 0) && (r_speed == 0)) timer10ms = 0;
			
			set_motor_speed(l_speed, r_speed);	
			
			if(option == 2){
					if((abs(r_speed) < 20) && (abs(l_speed) < 20)) timer10ms = 0;
					}
							
			
		}
		
		//Finished turning
		l_speed = r_speed = 0;				//Set motor speeds to zero
		set_motor_speed(l_speed, r_speed);
	
		
		
	
	
		
}

//-----------------------------------------------------------------------------

//Switch position = 2

void deadReckoning()
{   
	delay_start(500,deadrec);  // 5 sec delay
	refresh_sensors();
	int p = 75; //pot - 128;
	driveStraight(43, p);
	rotate(120,p,CW,0);
	driveStraight(48,p);
	rotate(120,p,CW,0);
	driveStraight(53,p);
	
	
	
	
}



//-----------------------------------------------------------------------------

// Switch position = 1

void beaconFinder()
{
	int speed, c;
	UCHAR near,stop;
	
	delay_start(500,beaconfinder);
	
	face_beacon();	
	
	near = FALSE;
	stop = FALSE;
	
	while(stop == FALSE){
		refresh_sensors();
		refreshIRdata();	
		speed = pot - 128;
		
		c = beacon.centroid * 2;   //Get beacon heading
		
		//Drive towards the beacon
		l_speed = speed + c;
		r_speed = speed - c;
		
		
		//Compute active IR obstacle steering
		c = activeIR.centroid ;
		
		if(c < 0){			//Split centroid into left and right "avoid" functions 	 
			c += 0x50; 
			if(c < 0) c = 0;
		}
		else{
			c -= 0x50;
			if(c > 0) c = 0;
		}
		
		c = c * 2;
		
		//Drive around obstacles if present
		if((activeIR.brightness > 30) && (near == FALSE)) {
						
			l_speed = speed - c;
			r_speed = speed + c;
						
		}
		
		
		set_motor_speed(l_speed,r_speed);   //Set motor speeds to computed values
		
		
		if((beacon.brightness < 0x48) && (near == FALSE)){  //If we lost sight of beacon spin until we find it again.
			face_beacon();			
			
		}
		
				
		if(beacon.brightness > 0xc0) near = TRUE;
		
		if((beacon.brightness < 0x50) && (near == TRUE)) stop = TRUE;
		
		
		
	}
	
	
	
}
	
	
	
//-----------------------------------------------------------------------------

// Switch = 0 

void lineFollow()
{   int p, n, q;
	int last_centroid, d , I,  pid;
	UCHAR mem_flag = LEFT;
	dir = FWD;
	delay_start(500,linefoll);  //5 sec delay
	run_seconds = 0;
	n = 0;
	q = 0;
	I = 0;
	last_centroid = 0;
	d = 0;
	
	while(run_seconds < 90){
		refreshLineData();
		p = line.centroid;
		refresh_sensors();
		speed = pot - 128;
				
		if(line.present == TRUE){  //If we see a line, track it
		
		p = p * 2;
		pid = p+d+(I/64);		
		r_speed = speed + pid;
		l_speed = speed - pid;
				
		} 
		else{  //Don't see a line, stop and spin to re-aquire line
			//I = 0;	
			if(mem_flag == LEFT){  //Spin in direction line last seen
				l_speed = -speed;  // CCW
				r_speed = speed;
			} else
				{ l_speed = speed;  // CW
				  r_speed = -speed;
				}
		}
		
		if(line.brightness < 0xd0){     //Go straight if over a intersecting black line
			l_speed = r_speed = speed;
		}
		
		set_motor_speed(l_speed, r_speed);  //Now set the motor speeds.
		//set_motor_speed(0,0);
		
		if(line.present == TRUE){
			if(line.map & 0xc0) mem_flag = LEFT;   //Remember line was on the left
			if(line.map & 0x03) mem_flag = RIGHT; // or the right
			}
		
		//Compute Derivative
		if(n++ >= 10){
			n = 0;
			d = line.centroid - last_centroid;
			d = d * 8;
			last_centroid = line.centroid;
		}
		
		//Compute integral
		if(q++ >= 2){
			q = 0;
			I = I + line.centroid;
			if(I > 4000) I = 4000;  //Limit range to +/- 4000
			if(I < -4000) I = -4000;
			}
			
			
		
		
		if(debugReady()){			
			debug[0] = I >> 8;
			debug[1] = I & 255;
			debug[2] = mem_flag;
			debug[3] = line.contrast;
			debug[4] =  line.centroid & 255;
			debug[5] =  line.map;
			refreshDebug();
			
		} 

	}
}

//--------------------------------------------------------------------------

//---------------------------------------------------------------------------

//Switch = 3

void bulldozer()
{   int sliceL, sliceR, floorL, floorR, tableL, tableR, c, r;
    UCHAR  d ;
	delay_start(500,bulldoz);  //5 sec delay
	
	run_seconds = 0;
	
	refresh_sensors();
	floorL = edge_left  ;  //get light level looking at the floor
	floorR = edge_right   ;
	
	
	trip_odo_left = CPI * 6;   //Setup to back up 6 inches
	while(trip_odo_left > 0) set_motor_speed(-80,-80);
	set_motor_speed(0,0);
	delay(20);
	refresh_sensors();
	tableL =  edge_left  ;  //get average table light level
	tableR =  edge_right  ;
	sliceL = floorL + ((tableL - floorL) / 3) ;   //compute value 33% up from dark
	sliceR = floorR + (( tableR - floorR) / 3) ;
	
	rotate(90,90,CW,2);		//rotate CW 90 degrees, option = fast and sloppy
	rotate(180,90,CW,1);  //Rotate 180 CW more or stop if object seen.
	rotate(90,90,CCW,1);  //Rotate CCW up to 90 if no object
	
	/*
	debug[0] = sliceL;
	debug[1] = floorL;
	debug[2] = tableL;
	debug[3] = sliceR;
	debug[4] = floorR;
	debug[5] = tableR;
	refreshDebug();
	*/
	
	while(run_seconds < 90){  //Limited to 90 seconds
		refreshLineData();
		refreshIRdata();
		refresh_sensors();		
		speed = 70; //pot - 128;
		
		c = activeIR.centroid ;   //Get object heading
		if(activeIR.brightness < 0x18) c = 0;   //If no object then centroid = 0
		
		l_speed = speed - c;	//Drive towards the object
		r_speed = speed + c;
		
		if((edge_right < sliceR) && (edge_left < sliceL)){   //Edge detected with both left and right sensors
			trip_odo_left = CPI * 3;
			while(trip_odo_left > 0) set_motor_speed(-80,-80); //Reverse 3 inches
			set_motor_speed(0,0);
			if(rnd & 1) d = CW ; else d = CCW;	//Select a random direction, CW or CCW
			rotate(90,80,d,2);		 //Rotate 90 degrees CW or CCW
			rotate(180,80,d,1);      //Rotate 180 scanning for objects
			d = d ^ 1;				 // reverse direction
			r = 27 + (rnd & 0x3f);
			rotate(r,80,d,1);		//Rotate 27 to 90 degrees randomly or until object seen
			
			}
			
		   //Edge detected with either edge sensor
		   //This causes the bot to square itself with the edge
		   refresh_sensors();
			if(edge_right < sliceR){  r_speed = -40; l_speed = 50; }
			if(edge_left < sliceL) { l_speed = -40;	r_speed = 50; }
		
				
		set_motor_speed(l_speed,r_speed);   //Set motor speeds to computed values
		
			
		
		}
	
	
	 
	
	
}

//----------------------------------------------------------------------------

/* Switch Position 4:  Line maze.

 Right hand follow rule.  No learning in this version.
 1. If dead end, turn around 180 degrees
 2. If choice is left or straight, go straight
 3. If choice is right or straight, go right.
 4. If choice is left or right, go right.
 
*/

void lineMaze(char rule)
{
    char noLine;
	int p, n, q, deg;
	
	long dest;
	
	char left_flag, right_flag;
	char turn = CW;
	char fat_line_counter = 0;
	dir = FWD;
	
	delay_start(500,linemaze);  //5 sec delay
	run_seconds = 0;
	n = 0;
	q = 0;
	noLine = 0;
	
		
	left_flag = right_flag = FALSE;
	timer_trip = 3 ; //30ms
	
	//Maze loop
	while(run_seconds < 90){  //Run for 90 seconds max.
		refreshLineData();
		p = line.centroid;
		refresh_sensors();
		speed = pot - 128;   //Pot sets speed.
		
						
		if((line.width > 0) && (line.width < 3)){  //If we see a normal thin line, track it
		
		p = p * 2;				
		
		r_speed = speed + p;		//Apply Proportional error term
		l_speed = speed - p;
				
		} else {l_speed = speed; r_speed = speed;}  //else fat line. Assume 90 deg corner or intersection. Don't track.
		
			
		/* Intersection detecton and Left or right turn process */
		if((edge_left < ET) || (edge_right < ET) || (line.present == FALSE)){
			set_motor_speed(60,60);				//Slow down a little and proceed straight.
			left_flag = right_flag = FALSE;		//reset intersection flags	
			dest = trip_odo_left + (CPI * 2);  //Move 2 inches past intersection			
			while(trip_odo_left < dest) {
				if(edge_left < ET) left_flag = TRUE;	//update intersection flags as we move	
				if(edge_right < ET) right_flag = TRUE;
			    refresh_sensors();
				refreshLineData();
				
				} 
				
					
				deg = 90;   //Assume 90 degree turn
				turn = NONE; //Default to straight ahead
			
				if(line.present == FALSE){
					if((left_flag == TRUE) && (right_flag == FALSE)) turn = CCW;  //Left turn corner	
					if((left_flag == FALSE) && (right_flag == TRUE)) turn = CW;  //Right turn corner
					
					if((left_flag == TRUE) && (right_flag == TRUE)){ // T intersection choice...
						if(rule == RIGHT) turn = CW;   //right
						if(rule == LEFT) turn = CCW;	//left
						}
						
					if((right_flag == FALSE) && (left_flag == FALSE)){  //Dead end, turn around
						turn = CW;
						deg = 180;
					}						
				}	
				
				if(line.present == TRUE){
					if((rule == RIGHT) && (right_flag == TRUE)) turn = CW;
					if((rule == LEFT) && (left_flag == TRUE)) turn = CCW;
													
				}
									
				
						
			if(turn != NONE) rotate(deg,80,turn,2);		 // Now do the turn	
			left_flag = FALSE;
			right_flag = FALSE;
			fat_line_counter = 0;   //Required to prevent false stops	
			noLine = 0;		
			l_speed = r_speed = speed;
			
			}
			
			
		/* End square detection */	
		if(timer_trip == 0){   //Every 20ms check line width to detect 4x4 inch end square
			timer_trip = 2;
			if(line.width > 3) fat_line_counter++;  // Count consecutive times we see a fat line.
				else fat_line_counter = 0;
			if(fat_line_counter > 4) run_seconds = 90;  // Yes, stop when we run into the 4x4 inch square
			
			if(line.present == FALSE) noLine++; else noLine = 0;
			
			} 
			
			
		set_motor_speed(l_speed, r_speed);  // set the motor speeds.
		
					
		

	}
}





		

//---------------------------------------------------------------------------

/* Switch Position F :  Debug Code to display Active IR object sensor values */

void test_sensors(void)
{	UCHAR adr = 0;

	unsigned int ctr,x;
	
		
	
	sbi(PORTC,3);		//Disable IR LEDs
	x = 0 ;
	
	while (1 == 1){
		
		irsensor[adr] = readActiveIR(adr);
		
		for(ctr=0;ctr<2560;ctr++){ sbi(PORTC,3); }  //About 1 ms between pulses
		adr++;					//Next LED address
		x++;
		
		if (x >= 100){  //about 10 updates per sec
			for(ctr=0;ctr < 6;ctr++) debug[ctr] = irsensor[ctr];
			
			refreshDebug();
			x = 0;
			}
		}
		
		



}

//--------------------------------------------------------------------------

//Switch = A
void test_ir()
{   int c;
	
	while(1 == 1){
		refreshIRdata();
	
		debug[0] = activeIR.centroid;
		debug[1] = activeIR.map ;
		debug[2] = activeIR.brightness;
		
		c = activeIR.centroid ;
		if(c < 0) c += 0x50; else c -= 0x50;
		debug[3] = c & 255;
		
		timer10ms = 15;
		while(timer10ms > 0);
		refreshDebug();
		//set_motor_speed(80,80);
	}
	
	
}




//---------------------------------------------------------------------

//Switch position = B

void test_beacon()
{
	
	refresh_sensors();
	while(1 == 1){
		refresh_sensors();
		debug[0] = beacon.left;
		debug[1] = beacon.right;
		debug[2] = beacon.brightness;
		debug[3] = beacon.centroid & 255;
		debug[4] = 0;
		debug[5] = 0;
		
		timer10ms = 15;
		while(timer10ms > 0);
		refreshDebug();
	}
	
	
	
}



//Switch position = E
/*
void test_passive_sensors()
{

		
	while(1 == 1){
		refreshLineData();
		debug[0] = get_adc8(AD_BEACON_LEFT);
		debug[1] = get_adc8(AD_BEACON_RIGHT);
		debug[2] = get_adc8(AD_POT);

		debug[3] = line.data[7];		
		debug[4] = line.data[0]	;	
		debug[5] = get_adc8(AD_MOTOR_CURRENT);
		
		timer10ms = 15;
		while(timer10ms > 0);
		refreshDebug();
		}

}

*/
//Switch position = D

void test_line_photocells()  
{  
	while(1 == 1){
		refreshLineData();
		refresh_sensors();
		debug[0] = line.brightness;  //Varies from 0x80 (black) to 0xd0 (white)
		debug[1] = edge_left;		//Table edge left
		debug[2] = edge_right;		
		debug[3] = line.contrast;
		debug[4] =  line.centroid & 255;
		debug[5] =  line.map;
		
		refreshDebug();
	}
}

/*
//Switch position 9
void test_first_6_photocells()
{ UCHAR i;
while(1 == 1){
		refreshLineData();
		refresh_sensors();
		for(i=0;i<6;i++) debug[i] = line.data[i];
		
		refreshDebug();
	}
}

*/
//--------------------------------------------------------

//Switch pos 8

//Save photocell white levels in EEPROM
//This must be done at least once.
//Bot must be on a white surface with no line!


int saveLineWhiteLevels()
{
    int i;
	UCHAR adr, cksum;
	
	delay_start(500,whitelevel);  //5 sec delay
	cksum = 0;
	//Read photocells.  Assume we are on pure white with no black line 
	for(adr=0;adr<8;adr++){
		if(adr & 1) sbi(PORTC,0); else cbi(PORTC,0);
		if(adr & 2) sbi(PORTC,1); else cbi(PORTC,1);
		if(adr & 4) sbi(PORTC,2); else cbi(PORTC,2);
		line.data[adr] = get_adc8(AD_LINE_FOLLOW);
		
		}
		
	line.data[8] = checksum(line.data,8);
	
     
   eeprom_write_block (line.data, (void *)0, 9) ; //Save values in EEPROM
   eeprom_write_block (line.data, (void *)16, 9) ; //Save values in EEPROM backup location
   
   eeprom_read_block((void*)&line.data, (const void*)0, 9);  //Read back for verification
   
   
   if(checksum(line.data,8) != line.data[8]) return -1;
      
   for(i=0;i<6;i++) debug[i] = line.data[i];
   refreshDebug();								//Show first 6 values on LCD
   run_seconds = 0;
   while(run_seconds < 5);
   return 0;
   


} 

//-----------------------------------------------------------


// Switch position = C

/*void test_quadrature()

{   int d;
	dir=FWD;
	odo_right = 0;
	odo_left = 0;
	delay_start(500,(PGM_P)testquad);  //5 sec delay
	timer_trip = 200;
	d = 0;
	
	while(1 == 1){
		refresh_sensors();
		speed = pot - 128;
		if(timer_trip == 0) speed = 0;
		l_speed = speed;
		r_speed = speed;
		d = odo_right - odo_left;
		if(d > 10) d = 10;
		if(d < -10) d = -10;
		
		l_speed = l_speed + d;
		r_speed = r_speed - d;
		set_motor_speed(l_speed, r_speed);
		
		if(debugReady()){
			debug[0] = (odo_right >> 16) & 255;
			debug[1] = (odo_right >> 8) & 255;
			debug[2] = (odo_right & 255);
			debug[3] = (odo_left >> 16) & 255;
			debug[4] = (odo_left >> 8) & 255;
			debug[5] = (odo_left & 255);
			refreshDebug();
		}
	}
		
		
	
} */

//---------------------------------------------------------------------------

int main(void)
{ 
  char *sbuf;
  int last_sec, rv;
 
/* Digital ports setup */
  outp(0x80, PORTD);          
  outp(0x32, DDRD);           /* port D inp/output setup */

  outp(0x0f,PORTB);           /* PORTB: low 4 bits pull ups for selector switch */
  outp(0x00,DDRB);            /* all bits inputs for octal selector switch and block sensor*/

  outp(0xff,PORTC);           // 
  outp(0xff,DDRC);            // All outputs on port C 

/* PORTA: A/D Converter setup */
  outp(0x00,DDRA);            // 8 bits analog inputs
  outp(0,PORTA);              // Clear port A
  outp(0x86,ADCSRA);           // Enable A to D converter with 16mhz/64 conv clk.  ( 250 KHZ)

/* Timer 1 dual PWM setup */
  outp(0x01,TCCR1B);            // clk/1 prescale, 8 bit phase correct pwm  (28.8 khz pwm freq))
  outp(0xa1,TCCR1A);            // 8 bit pwm, OCnx clear at top
  outp(0,OCR1AH);               //Clear the output compare registers
  outp(0,OCR1AL);
  outp(0,OCR1BH);
  outp(0,OCR1BL);

/* Timer-counter 1 setup */
  outp(BV(CS01), TCCR0);      /* use CLK/8 source for counter */


/* Timer-Counter 2 setup */
  outp(0x0b,TCCR2);           // clear on compare, Prescale = 32 on T/C 2  (500 khz with 16mhz osc)
  outp(49,OCR2);              // T/C 2 compare register for 10,000 hz interrupts (clk/(32*(1+OCR2))
  sbi(TIMSK,OCIE2);           // Enable timer 2 compare interrupts

/* UART setup */
  outp(103,UBRRL);              /* UART baud rate set to 9600 (16mhz osc) */
  sbi(UCSRB,TXEN);              // enable uart transmitter
	
/*External interrupt setup*/
	MCUCR = 0x0a;				//Falling edge interrupts on INT0 and INT1
	sbi(GICR,INT1);				//Enable INT1
	sbi(GICR,INT0);				//Enable INT0
	

  trip_counter = 0;
  txstate = 0;
  minutes = 0;
  run_seconds = 0;
  last_sec = 0xffff;
  last_sec = 0;
  
  odo_left = 0;
  odo_right = 0;
  line_history_odo_right = 0;
  line_history_odo_left = 0;
  
  l_speed = 0;
  r_speed = 0;
  timer10ms = 500;				//Set for 5 second delay
  timer_overcurrent = 100;		//1000ms
  sel_sw = (PINB ^ 0x0f) & 0x0f;		//Read selector switch
  
  
  sei();                      /* enable global interrupts */
  
 // wdt_enable(WDTO_250MS);          /* enable watchdog timer - 250 ms timeout. */ 
	
	cbi(PORTC,7);	//Test LED off
    sbi(PORTC,3);   //IR LEDs off
	cbi(PORTB,4);   //Disable motors
	set_motor_speed(0,0);
	dir = FWD;
	
  delay_10ms(10);			//Wait 100ms for LCD to init.	
  
 
  rv = 0;
  
  switch (sel_sw){			//Run job based on selector switch position
	  case 0:	lineFollow(); break;
	  case 1:	beaconFinder(); break;
	  case 2:	deadReckoning(); break;
	  case 3:	bulldozer(); break;
	  case 4:	lineMaze(RIGHT); break;
	  case 5:	lineMaze(LEFT); break;

	  case 0x0f:	test_sensors(); break; 
	  //case 0x0e:   test_passive_sensors(); break; 
	  case 0x0d:	test_line_photocells(); break;
	 // case 0x0c:	test_quadrature(); break;
	  case 0x0b:   test_beacon();  break;
	  case 0x0a:	test_ir(); break;
	  //case 0x09:    test_first_6_photocells(); break;
	  
	  case 0x08:	rv = saveLineWhiteLevels(); break;    //White level calibration
  
  };
  
 r_speed = l_speed = 0; 
 set_motor_speed(l_speed, r_speed);
 cbi(PORTB,4);   //Disable motors
 
 if(rv == 0) sbuf = "\x0cJob\r\nComplete";
	else sbuf = "\x0cError";
 writelcd(sbuf);
 while(1 == 1);		//Stop here.  Infinite loop.
 return 0;

};

/* End of File */
